import hugr.model

def term_to_string(term: hugr.model.Term) -> str: ...
def string_to_term(string: str) -> hugr.model.Term: ...
def node_to_string(node: hugr.model.Node) -> str: ...
def string_to_node(string: str) -> hugr.model.Node: ...
def region_to_string(region: hugr.model.Region) -> str: ...
def string_to_region(string: str) -> hugr.model.Region: ...
def param_to_string(region: hugr.model.Param) -> str: ...
def string_to_param(string: str) -> hugr.model.Param: ...
def symbol_to_string(region: hugr.model.Symbol) -> str: ...
def string_to_symbol(string: str) -> hugr.model.Symbol: ...
def module_to_string(module: hugr.model.Module) -> str: ...
def string_to_module(string: str) -> hugr.model.Module: ...
def module_to_bytes(module: hugr.model.Module) -> bytes: ...
def bytes_to_module(binary: bytes) -> hugr.model.Module: ...
def package_to_string(package: hugr.model.Package) -> str: ...
def string_to_package(string: str) -> hugr.model.Package: ...
def package_to_bytes(package: hugr.model.Package) -> bytes: ...
def bytes_to_package(binary: bytes) -> hugr.model.Package: ...
